BASE_URL = 'https://home.personalcapital.com'
API_ENDPOINT = BASE_URL + '/api'
STAPLE_CATEGORIES = [
    # 'Personal Care',
    'Gasoline/Fuel',
    'Groceries',
    'Utilities',
    'Healthcare/Medical',
    'Rent',
    'Pets/Pet Care',
    'Education',
    'Cable/Satellite',
    'Insurance',
    'Loans',
    'Wedding'
]



import re
import os
import json
import logging
import getpass
import requests
from os import path

from .exceptions import (
    RequireTwoFactorException,
    LoginFailedException
)

from .constants import (
    BASE_URL,
    API_ENDPOINT
)


class AuthLevelEnum(object):
    USER_REMEMBERED = "USER_REMEMBERED"


class TwoFactorVerificationModeEnum(object):
    SMS = 0
    # PHONE = 1
    EMAIL = 2


class Connector(object):

    # cross-site request forgery token params
    CSRF_KEY = "csrf"
    CSRF_TOKEN_REGEXP = re.compile(r"globals.csrf='([a-f0-9-]+)'")

    AUTH_LEVEL_KEY = "authLevel"
    SUCCESS_KEY = "success"
    SP_HEADER_KEY = "spHeader"
    ERRORS_KEY = "errors"

    def __init__(self):
        self.__session = CookieSession(API_ENDPOINT)

    def login(self, username, password):
        initial_csrf = self.__get_csrf_from_home_page(BASE_URL)
        csrf, auth_level = self.__identify_user(username, initial_csrf)

        if csrf and auth_level:
            self.session.csrf = csrf
            if auth_level != AuthLevelEnum.USER_REMEMBERED:
                raise RequireTwoFactorException()

            result = self.__authenticate_password(password).json()
            if self.__get_header_value(result, Connector.SUCCESS_KEY) is False:
                raise LoginFailedException(self.__get_error_value(result))
        else:
            raise LoginFailedException("Did not acquire CSRF token and auth level.")

    def authenticate_password(self, password):
        return self.__authenticate_password(password)

    def two_factor_authenticate(self, mode, code):
        if mode == TwoFactorVerificationModeEnum.SMS:
            return self.__authenticate_sms(code)
        elif mode == TwoFactorVerificationModeEnum.EMAIL:
            return self.__authenticate_email(code)

    def two_factor_challenge(self, mode):
        if mode == TwoFactorVerificationModeEnum.SMS:
            return self.__challenge_sms()
        elif mode == TwoFactorVerificationModeEnum.EMAIL:
            return self.__challenge_email()

    @property
    def csrf(self):
        return self.session.csrf

    @csrf.setter
    def csrf(self, csrf):
        self.session.csrf = csrf

    @property
    def session(self):
        return self.__session

    @classmethod
    def connect(cls):
        pc = cls()
        user, pw = cls.get_user_credentials()

        try:
            pc.login(user, pw)
        except RequireTwoFactorException:
            pc.two_factor_challenge(TwoFactorVerificationModeEnum.SMS)
            pc.two_factor_authenticate(TwoFactorVerificationModeEnum.SMS, input('SMS code: '))
            pc.authenticate_password(pw)

        pc.session.csrf = pc.csrf

        return pc.session

    @staticmethod
    def get_user_credentials():
        email = os.getenv('PEW_EMAIL')
        password = os.getenv('PEW_PASSWORD')

        if not email:
            print("You can set the environment variables for PEW_EMAIL and PEW_PASSWORD so "
                  "the prompts don't come up every time")
            return input('Enter email:')

        if not password:
            return getpass.getpass('Enter password:')

        return email, password

    @classmethod
    def __get_header_value(cls, result, valueKey):
        if (cls.SP_HEADER_KEY in result) and (valueKey in result[cls.SP_HEADER_KEY]):
            return result[cls.SP_HEADER_KEY][valueKey]
        return None

    @classmethod
    def __get_error_value(cls, result):
        try:
            return cls.__get_header_value(result, cls.ERRORS_KEY)[0]['message']
        except (ValueError, IndexError):
            return None

    def __get_csrf_from_home_page(self, url):
        r = self.__session.session.get(url)
        found_csrf = Connector.CSRF_TOKEN_REGEXP.search(r.text)

        if found_csrf:
            return found_csrf.group(1)
        return None

    def __identify_user(self, username, csrf):
        """
        Returns reusable CSRF code and the auth level as a 2-tuple
        """
        data = {
            "username": username,
            "csrf": csrf,
            "apiClient": "WEB",
            "bindDevice": "false",
            "skipLinkAccount": "false",
            "redirectTo": "",
            "skipFirstUse": "",
            "referrerId": "",
        }

        r = self.session.post("/login/identifyUser", data)

        if r.status_code == requests.codes.ok:
            result = r.json()
            new_csrf = self.__get_header_value(result, Connector.CSRF_KEY)
            auth_level = self.__get_header_value(result, Connector.AUTH_LEVEL_KEY)
            return (new_csrf, auth_level)

        return (None, None)

    def __generate_challenge_payload(self, challenge_type):
        return {
            "challengeReason": "DEVICE_AUTH",
            "challengeMethod": "OP",
            "challengeType": challenge_type,
            "apiClient": "WEB",
            "bindDevice": "false",
            "csrf": self.csrf
        }

    def __generate_authentication_payload(self, code):
        return {
            "challengeReason": "DEVICE_AUTH",
            "challengeMethod": "OP",
            "apiClient": "WEB",
            "bindDevice": "false",
            "code": code,
            "csrf": self.csrf
        }

    def __challenge_email(self):
        data = self.__generate_challenge_payload("challengeEmail")
        return self.session.post("/credential/challengeEmail", data)

    def __authenticate_email(self, code):
        data = self.__generate_authentication_payload(code)
        return self.session.post("/credential/authenticateEmail", data)

    def __challenge_sms(self):
        data = self.__generate_challenge_payload("challengeSMS")
        return self.session.post("/credential/challengeSms", data)

    def __authenticate_sms(self, code):
        data = self.__generate_authentication_payload(code)
        return self.session.post("/credential/authenticateSms", data)

    def __authenticate_password(self, passwd):
        data = {
            "bindDevice": "true",
            "deviceName": "",
            "redirectTo": "",
            "skipFirstUse": "",
            "skipLinkAccount": "false",
            "referrerId": "",
            "passwd": passwd,
            "apiClient": "WEB",
            "csrf": self.csrf
        }
        return self.session.post("/credential/authenticatePassword", data)


class CookieSession(object):

    SESSION_FN = path.join(path.expanduser("~"), ".personalcapital", "session.json")

    def __init__(self, api_endpoint):
        self.__session = requests.Session()
        self.__session_file = CookieSession.SESSION_FN
        self.__api_endpoint = api_endpoint
        self.__csrf = ""
        self.load()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.save()

    @property
    def api_endpoint(self):
        return self.__api_endpoint

    @api_endpoint.setter
    def api_endpoint(self, new_endpoint):
        self.__api_endpoint = new_endpoint

    @property
    def session(self):
        return self.__session

    @session.setter
    def session(self, new_session):
        raise NotImplementedError("Re-assigning internal session not supported.")

    @property
    def cookies(self):
        return requests.utils.dict_from_cookiejar(self.session.cookies)

    @cookies.setter
    def cookies(self, cookies):
        self.session.cookies = requests.utils.cookiejar_from_dict(cookies)

    @property
    def csrf(self):
        return self.__csrf

    @csrf.setter
    def csrf(self, csrf):
        self.__csrf = csrf

    def fetch(self, endpoint, data=None):
        """
        for getting data after logged in
        """
        payload = {
            "lastServerChangeId": "-1",
            "csrf": self.csrf,
            "apiClient": "WEB"
        }
        if data is not None:
            payload.update(data)

        return self.post(endpoint, payload)

    def post(self, endpoint, data):
        return self.session.post(self.api_endpoint + endpoint, data)

    def load(self, session_file=None):
        fn = session_file or self.SESSION_FN
        if path.isfile(fn):
            with open(fn) as data_file:
                cookies = {}
                try:
                    cookies = json.load(data_file)
                except ValueError as err:
                    logging.error(err)
                self.cookies = cookies
        return self

    def save(self):
        with open(self.__session_file, 'w') as data_file:
            data_file.write(json.dumps(self.cookies))

    def make_request(self, endpoint):
        response = self.fetch(endpoint)

        # throw error if bad, parse json if good
        if response.status_code != 200:
            raise ValueError("Request failed with code {}.".format(response.status_code))
        return response.json()['spData']



from personalcapital import (
    Connector,
    Database
)
from personalcapital.etl import get_expenses, get_accounts
from datetime import datetime, date
from .log import logger


def current_expense_margin(month=None):
    df = get_expenses(month)
    non_staple = df.loc[(~df.is_staple) & (df.isSpending)]
    return 3200 - non_staple['amount'].sum()


def get_report():
    monthly_expense_margin = current_expense_margin()
    accounts = get_accounts()
    group_sums = accounts.groupby('accountTypeGroup').sum()['balance']
    net_worth = accounts['balance'].sum()
    loan_debt = group_sums.loc['LOAN']
    car_loan_debt = accounts.loc[accounts['firmName'].str.contains('Toyota'), 'balance'].sum()
    student_loan_debt = accounts.loc[accounts['firmName'].str.contains('Navient'), 'balance'].sum()
    credit_debt = group_sums.loc['CREDIT_CARD']
    retirement = group_sums.loc['RETIREMENT']
    investments = group_sums.loc['INVESTMENT']
    bank = group_sums.loc['BANK']

    return {
        "1) Net Worth": "{:.2f}".format(net_worth),
        "2) Loans (All)": "{:.2f}".format(loan_debt),
        "3) Loans (Auto)": "{:.2f}".format(car_loan_debt),
        "4) Loans (Student)": "{:.2f}".format(student_loan_debt),
        "5) Credit Cards": "{:.2f}".format(credit_debt),
        "6) Bank": "{:.2f}".format(bank),
        "7) Retirement": "{:.2f}".format(retirement),
        "8) Monthly Margin": "{:.2f}".format(monthly_expense_margin),
        "9) Last Update": datetime.now().strftime("%m/%d %H:%M")
    }



import pytest
import pandas as pd

from personalcapital.etl import (
    get_category_map,
    get_accounts
)


def test_basic():
    cmap = get_category_map()
    assert isinstance(cmap, dict)

    accounts = get_accounts()
    assert isinstance(accounts, pd.DataFrame)



import json
import pandas as pd
from datetime import datetime, date, timedelta

from .log import logger
from personalcapital import (
    Connector,
    Database
)
from personalcapital.constants import (
    STAPLE_CATEGORIES
)


def get_category_map():
    with Connector.connect() as session:
        category_map = {
            item['transactionCategoryId']: item['name']
            for item in session.make_request("/transactioncategory/getCategories")
        }
    return category_map


def get_transactions():
    with Database() as db:
        txns = db.get_transactions()
    return txns


def update_transactions():
    def _update_transactions(session):
        now = datetime.now()
        date_format = '%Y-%m-%d'
        days = 365 * 2
        start_date = (now - (timedelta(days=days+1))).strftime(date_format)
        end_date = (now - (timedelta(days=1))).strftime(date_format)
        txn_response = session.fetch('/transaction/getUserTransactions', {
            'sort_cols': 'transactionTime',
            'sort_rev': 'true',
            'page': '0',
            'rows_per_page': '100',
            'startDate': start_date,
            'endDate': end_date,
            'component': 'DATAGRID'
        })

        # throw error if bad, parse json if good
        if txn_response.status_code != 200:
            raise ValueError("Fetching transactions failed with code {}.".format(txn_response.status_code))
        txn_response = txn_response.json()

        # parse the data out and return
        transactions = txn_response['spData']
        print('Number of transactions between {0} and {1}: {2}'.format(transactions['startDate'], transactions['endDate'],
                                                                       len(transactions['transactions'])))

        return transactions['transactions']

    with Connector.connect() as session, Database() as db:
        txns = _update_transactions(session)
        db.clear_transactions()
        db.add_transactions(txns)


def get_expenses(month=None):
    if month is None:
        month = date.today().strftime("%Y-%m")
    df = pd.DataFrame(get_transactions())

    # filter by the month of interest
    df['year_month'] = df['transactionDate'].map(lambda d: '-'.join(d.split('-')[:2]))
    df = df.loc[df['year_month'] == month]

    cmap = get_category_map()
    df['category'] = df['categoryId'].map(cmap)
    df['is_staple'] = df['category'].map(lambda cat: cat in STAPLE_CATEGORIES)
    df['transactionDate'] = pd.to_datetime(df['transactionDate'])
    return df


def get_accounts():
    with Connector.connect() as session:
        accounts = pd.DataFrame(session.make_request('/newaccount/getAccounts')['accounts'])
    accounts.loc[accounts.isLiability, 'balance'] = -1 * accounts.loc[accounts.isLiability, 'balance']
    return accounts



from pymongo import MongoClient
from .log import logger


class Database(object):
    def __init__(self):
        self.client = MongoClient('localhost', port=27017)
        self.db = self.client.personalcapital
        self.txns = self.db.transactions
        # self.accounts = self.db.accounts

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def add_transactions(self, txns):
        existing_ids = set([
            e['userTransactionId']
            for e in self.txns.find({}, ["userTransactionId"])
        ])
        passed_ids = set([
            e['userTransactionId']
            for e in txns
        ])
        new_ids = passed_ids - existing_ids

        new_txns = [
            txn
            for txn in txns
            if txn['userTransactionId'] in new_ids
        ]

        if len(new_txns) > 0:
            logger.info("Adding {} new transactions.".format(len(new_txns)))
            self.txns.insert_many(new_txns)
        else:
            logger.info("No new transactions to add to the database.")

    def get_transactions(self):
        return list(self.txns.find())

    def clear_transactions(self):
        self.txns.remove({})



import os
import json
import numpy as np
import pandas as pd
from os import path
import matplotlib.pyplot as plt
from argparse import (
    ArgumentParser,
    ArgumentDefaultsHelpFormatter
)
from collections import defaultdict, Counter

from requests import Session
from personalcapital.analyze import get_report
from personalcapital.etl import update_transactions


def report():
    if "PEW_PAYLOAD_URL" not in os.environ:
        raise EnvironmentError("PEW_PAYLOAD_URL is not set.")

    report = get_report()
    session = Session()
    response = session.post(os.environ["PEW_PAYLOAD_URL"], json=report)
    print(json.dumps(report, indent=4))


def get_clargs():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(dest="action", choices=['report', 'update'], help="Action to perform.")
    return parser.parse_args()


def main():

    def _main(action):
        if action == 'update':
            update_transactions()
        elif action == 'report':
            report()
        else:
            raise ValueError("Unsupported action '{}'".format(action))

    args = get_clargs()
    _main(args.action)



import json
from os import path
from glob import glob
from setuptools import setup, find_packages


def get_package_info():
    with open("setup.json", 'r') as fp:
        cfg = json.load(fp)
    return cfg["package_info"]


def get_requirements():
    requirements_fn = path.join(path.dirname(__file__), 'requirements.txt')
    with open(requirements_fn, 'r') as fh:
        requirements = [str(x).strip() for x in fh.readlines()]
    return requirements


def get_version(module_name):
    version_fn = path.join(module_name, "version.py")
    with open(version_fn) as fp:
        version = float(fp.readlines()[0].strip().split('=')[1].replace(' ', ''))
    return version


def get_data_files():
    data_files = ['README.rst']
    if path.isdir("conf"):
        data_files.extend(glob(path.join("conf", "*")))
    return data_files


def get_entry_points():
    with open("setup.json") as fp:
        cfg = json.load(fp)
    return cfg.get("entry_points", {})


package_info = get_package_info()
module_name = package_info["module_name"]
version = get_version(module_name)

setup(
    name=package_info["friendly_name"],
    version=version,
    description=package_info["description"],
    long_description=package_info.get("long_description", package_info["description"]),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    url=package_info["url"],
    author=package_info["author"],
    author_email=package_info["author_email"],
    packages=find_packages(exclude=['*.test', 'test']),
    install_requires=get_requirements(),
    entry_points=get_entry_points(),
    zip_safe=False,
    data_files=[('share/{}'.format(module_name), get_data_files())]
)
